<?php
/*
Plugin Name: Dynamic Front-End Heartbeat Control
Description: An enhanced solution for optimizing the performance of your WordPress website when the default manual heartbeat settings or the WordPress Heartbeat API fall short. Say goodbye to guesswork and inefficiencies caused by default heartbeat settings.
Version: 1.1.9
Author: Codeloghin
Author URI: https://www.fiverr.com/codeloghin
License: GPL2
*/

define('DFEHC_MIN_INTERVAL', 15);
define('DFEHC_MAX_INTERVAL', 300);
define('DFEHC_MAX_SERVER_LOAD', 85);
define('DFEHC_MAX_RESPONSE_TIME', 5000);
define('SERVER_LOAD_FACTOR', 1);
$nonce = isset($_POST['nonce']) ? sanitize_text_field($_POST['nonce']) : '';
$server_load = isset($_POST['serverLoad']) ? floatval(sanitize_text_field($_POST['serverLoad'])) : 0;
$server_response_time = isset($_POST['serverResponseTime']) ? intval(sanitize_text_field($_POST['serverResponseTime'])) : 0;

function set_default_last_activity_time($user_id) {
    update_user_meta($user_id, 'last_activity_time', current_time('timestamp'));
}
add_action('user_register', 'set_default_last_activity_time');

class UserActivityProcess {
    private $queue;

    public function __construct() {
        $this->queue = new SplQueue(); 
    }

    public function push_to_queue($user) {
        $this->queue->enqueue($user);
    }

    public function process_queue() {
        while (!$this->queue->isEmpty()) {
            $user = $this->queue->dequeue();
            
            $lastActivity = get_user_meta($user->ID, 'last_activity_time', true);
            if (!$lastActivity) {
                error_log("Failed to fetch last activity time for user: {$user->ID}. This can be caused by recently cleared cache. Recalculating...");
                update_user_meta($user->ID, 'last_activity_time', current_time('timestamp'));
            }
        }
    }
}

function process_user_activity() {
    $user_activity_process = new UserActivityProcess();

    $batch_size = 75;
    $offset = 0;

    while ($users = get_users_in_batches($batch_size, $offset)) {
        foreach ($users as $user) {
            $user_activity_process->push_to_queue($user);
        }

        $offset += $batch_size;
    }

    $user_activity_process->process_queue();
}
add_action('init', 'process_user_activity');
add_action('dfehc_process_user_activity', 'process_user_activity');

function get_users_in_batches($batch_size, $offset) {
    $args = array(
        'number' => $batch_size,
        'offset' => $offset,
    );
    $user_query = new WP_User_Query($args);
    return $user_query->get_results();
}

function record_user_activity() {
    if (is_user_logged_in()) {
        $current_user = wp_get_current_user();
        $time = current_time('timestamp');
        update_user_meta($current_user->ID, 'last_activity_time', $time);
    }
}
add_action('wp_footer', 'record_user_activity');
add_action('wp', 'record_user_activity');

function schedule_user_activity_processing() {
    $user_activity_process = new UserActivityProcess();

    $batch_size = 75;
    $offset = 0;
    $users = array();

    do {
        $users = get_users_in_batches($batch_size, $offset);

        if ($users) {
            foreach ($users as $user) {
                $user_activity_process->push_to_queue($user);
            }
        }

        $offset += $batch_size;
    } while (count($users) === $batch_size);

    wp_schedule_single_event(time() + 10, 'dfehc_process_user_activity');
	 if (!wp_next_scheduled('dfehc_process_user_activity')) { wp_schedule_event(time(), 'hourly', 'dfehc_process_user_activity'); }
}
add_action('init', 'schedule_user_activity_processing');
add_action('dfehc_process_user_activity', 'process_user_activity');

function cleanup_user_activity() {
    $users = get_users();

    foreach ($users as $user) {
        delete_user_meta($user->ID, 'last_activity_time');
    }
}

if (!wp_next_scheduled('dfehc_cleanup_user_activity')) {
    wp_schedule_event(time(), 'daily', 'dfehc_cleanup_user_activity');
}

add_action('dfehc_cleanup_user_activity', 'cleanup_user_activity');

function dfehc_get_server_load() {
    $load_logs = get_option('dfehc_server_load_logs', array());
    $server_load = 0;

    if (function_exists('sys_getloadavg')) return sys_getloadavg()[0];
    if (function_exists('exec') && !empty($_SERVER['SERVER_ADDR'])) {
        preg_match('/load average: ([0-9.]+)/', implode(" ", exec("uptime", $output=[])), $matches);
        return !empty($matches[1]) ? (float) $matches[1] : 0;
    }
    if (is_readable('/proc/loadavg')) return (float) explode(' ', file_get_contents('/proc/loadavg'))[0];
    if (function_exists('shell_exec')) {
        preg_match('/load average: ([0-9.]+)/', shell_exec("uptime"), $matches);
        return !empty($matches[1]) ? (float) $matches[1] : 0;
    }
    if (is_readable('/proc/stat')) {
        preg_match('/^cpu\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)/', file_get_contents('/proc/stat'), $matches);
        return isset($matches[1], $matches[2], $matches[3], $matches[4]) ? (1 - ($matches[4] / ($matches[1] + $matches[2] + $matches[3] + $matches[4]))) * 100 : 0;
    }
    if (function_exists('microtime')) {
        $startTime = microtime(true);
        $endTime = $startTime + 0.1;
        $busyTime = 0;
        while (microtime(true) < $endTime) $busyTime++;
        return ($busyTime / ($busyTime + ($endTime - $startTime))) * 100;
    }

    $load_logs[] = array(
        'timestamp' => time(),
        'load' => $server_load,
    );
    while (end($load_logs)['timestamp'] < time() - 86400) {
        array_shift($load_logs);
    }
    update_option('dfehc_server_load_logs', $load_logs);

    return $server_load;
}

function get_server_health_status($load) {
    if ($load < 14.95) {
        return 'Resting';
    } elseif ($load >= 15.00 && $load <= 33.00) {
        return 'Pacing';
    } elseif ($load > 35.00 && $load <= 67.65) {
        return 'Under Load';
    } else {
        return 'Under Strain';
    }
}

function get_server_response_time() {
    $ping_command = '';
    $response_time = null;
    if (stristr(PHP_OS, 'win')) {
        $ping_command = 'ping -n 1 localhost';
    } else {
        $ping_command = 'ping -c 1 localhost';
    }
    $descriptors = array(
        0 => array('pipe', 'r'),
        1 => array('pipe', 'w'),
        2 => array('pipe', 'w')  
    );
    $process = proc_open($ping_command, $descriptors, $pipes);
    if (is_resource($process)) {
        stream_set_blocking($pipes[1], 0);
        stream_set_blocking($pipes[2], 0);
        $status = proc_get_status($process);
        while ($status['running']) {
            usleep(1000); 
            $status = proc_get_status($process);
        }
        $output = stream_get_contents($pipes[1]);
        fclose($pipes[1]);
        $time_regex = '/time=([0-9.]+) ms/';
        preg_match($time_regex, $output, $matches);

        if (!empty($matches[1])) {
            $response_time = (float) $matches[1];
        }
        fclose($pipes[0]);
        fclose($pipes[2]);
        proc_close($process);
    }
    return $response_time;
}
function load_average() {
    if (function_exists('sys_getloadavg')) {
        $loadavg = sys_getloadavg();
        return $loadavg[0];
    }
    return 0.0;
}

function heartbeat_health_dashboard_widget_function() {
    $heartbeat_status = get_transient('heartbeat_health_status');
    $server_load = dfehc_get_server_load();
    $server_response_time = get_server_response_time(); 
    if ($heartbeat_status === false) {
        $heartbeat_status = get_server_health_status($server_load);
        set_transient('heartbeat_health_status', $heartbeat_status, 60);
    }
    $recommended_interval = calculate_recommended_interval_user_activity($server_load);
    $load_logs = get_option('dfehc_server_load_logs', array());
    $load_logs[] = array(
        'timestamp' => time(),
        'load' => $server_load,
    );
    while (end($load_logs)['timestamp'] < time() - 86400) {
        array_shift($load_logs);
    }
    update_option('dfehc_server_load_logs', $load_logs);

    if ($heartbeat_status === false) {
        $heartbeat_status = get_server_health_status($server_load);
        set_transient('heartbeat_health_status', $heartbeat_status, 60);
    }
    
        switch ($heartbeat_status) {
        case 'Resting':
            $status_color = 'green';
            $animation_name = 'heartbeat-healthy';
            break;
        case 'Pacing':
            $status_color = 'green';
            $animation_name = 'heartbeat-under-load';
            break;
        case 'Under Load':
            $status_color = 'yellow';
            $animation_name = 'heartbeat-working-hard';
            break;
        case 'Under Strain':
            $status_color = 'red';
            $animation_name = 'heartbeat-critical';
            break;
    }

    echo "<style>
        .heartbeat {
            animation: {$animation_name} 1s linear infinite;
        }

        @keyframes heartbeat-healthy {
            0% {box-shadow: 0 0 5px {$status_color}, 0 0 10px {$status_color}, 0 0 15px {$status_color}, 0 0 20px {$status_color};}
            50% {box-shadow: 0 0 20px {$status_color}, 0 0 30px {$status_color}, 0 0 40px {$status_color}, 0 0 50px {$status_color};}
            100% {box-shadow: 0 0 5px {$status_color}, 0 0 10px {$status_color}, 0 0 15px {$status_color}, 0 0 20px {$status_color};}
        }
      @keyframes heartbeat-under-load {
            0% {box-shadow: 0 0 5px {$status_color}, 0 0 10px {$status_color}, 0 0 15px {$status_color}, 0 0 20px {$status_color};}
            25% {box-shadow: 0 0 30px {$status_color}, 0 0 40px {$status_color}, 0 0 50px {$status_color}, 0 0 60px {$status_color};}
            50% {box-shadow: 0 0 5px {$status_color}, 0 0 10px {$status_color}, 0 0 15px {$status_color}, 0 0 20px {$status_color};}
            75% {box-shadow: 0 0 30px {$status_color}, 0 0 40px {$status_color}, 0 0 50px {$status_color}, 0 0 60px {$status_color};}
            100% {box-shadow: 0 0 5px {$status_color}, 0 0 10px {$status_color}, 0 0 15px {$status_color}, 0 0 20px {$status_color};}
        }
        @keyframes heartbeat-working-hard {
            0% {box-shadow: 0 0 5px {$status_color}, 0 0 10px {$status_color}, 0 0 15px {$status_color}, 0 0 20px {$status_color};}
            50% {box-shadow: 0 0 20px {$status_color}, 0 0 30px {$status_color}, 0 0 40px {$status_color}, 0 0 50px {$status_color};}
            100% {box-shadow: 0 0 5px {$status_color}, 0 0 10px {$status_color}, 0 0 15px {$status_color}, 0 0 20px {$status_color};}
        }
        @keyframes heartbeat-critical {
            0% {box-shadow: 0 0 5px {$status_color}, 0 0 10px {$status_color}, 0 0 15px {$status_color}, 0 0 20px {$status_color};}
            25% {box-shadow: 0 0 20px {$status_color}, 0 0 30px {$status_color}, 0 0 40px {$status_color}, 0 0 50px {$status_color};}
            50% {box-shadow: 0 0 5px {$status_color}, 0 0 10px {$status_color}, 0 0 15px {$status_color}, 0 0 20px {$status_color};}
            75% {box-shadow: 0 0 20px {$status_color}, 0 0 30px {$status_color}, 0 0 40px {$status_color}, 0 0 50px {$status_color};}
            100% {box-shadow: 0 0 5px {$status_color}, 0 0 10px {$status_color}, 0 0 15px {$status_color}, 0 0 20px {$status_color};}
        }
    </style>";
    $load_logs = get_option('dfehc_server_load_logs', array());
    $load_logs[] = array(
        'timestamp' => time(),
        'load' => $server_load,
    );
    while (end($load_logs)['timestamp'] < time() - 86400) {
        array_shift($load_logs);
    }
    update_option('dfehc_server_load_logs', $load_logs);
    echo "<p style='text-align: center; font-size: 24px; margin-top: 20px;'>Systolic Pressure: <strong>{$server_load}</strong></p>";
    echo "<div style='width: 30px; height: 30px; background-color: {$status_color}; margin: 20px auto; border-radius: 50%; animation: {$animation_name} 2s infinite;'></div>";
    echo "<p style='text-align: center; font-size: 24px; margin-top: 20px;'>Heartbeat: <strong>{$heartbeat_status}</strong></p>";
    $load_logs = get_option('dfehc_server_load_logs', array());
    $load_average = load_average();
    $server_load_factor = SERVER_LOAD_FACTOR - ($load_average / DFEHC_MAX_SERVER_LOAD);
$labels = array();
$data = array();
$timestamp = strtotime('-24 hours'); 
$interval = 20 * 60; 
while ($timestamp <= time()) {
    $load = 0;
    foreach ($load_logs as $log) {
        if ($log['timestamp'] >= $timestamp && $log['timestamp'] < ($timestamp + $interval)) {
            $load += $log['load'];
        }
    }
    $average_load = $load / count($load_logs);
    $labels[] = date('H:i', $timestamp);
    $data[] = $average_load;
    $timestamp += $interval; 
}
echo '<canvas id="loadChart"></canvas>';
    echo '
    <script src="https://cdn.jsdelivr.net/npm/chart.js"></script>
    <script>
    document.addEventListener("DOMContentLoaded", function() {
        var ctx = document.getElementById("loadChart").getContext("2d");
        var myChart = new Chart(ctx, {
            type: "line",
            data: {
                labels: ' . json_encode($labels) . ',
                datasets: [{
                    label: "Diastolic pressure",
                    data: ' . json_encode($data) . ',
                    backgroundColor: "rgba(75,192,192,0.2)",
                    borderColor: "rgba(75,192,192,1)",
                    borderWidth: 2
                }]
            },
           options: {
    responsive: true,
    scales: {
        y: {
            beginAtZero: true,
            max: 10,
            suggestedMax: 36, 
            ticks: {
                stepSize: 1 
            }
        }
    }
}
        });
    });
    </script>
    ';
}

function add_heartbeat_health_dashboard_widget() {
    wp_add_dashboard_widget('heartbeat_health_dashboard_widget', 'Dynamic Heartbeat Health Check', 'heartbeat_health_dashboard_widget_function');
}

add_action('wp_dashboard_setup', 'add_heartbeat_health_dashboard_widget' );

function dfehc_enqueue_scripts() {
    wp_enqueue_script('heartbeat', plugin_dir_url(__FILE__) . 'js/heartbeat.js', array('jquery'), '1.0', true);
    wp_script_add_data('heartbeat', 'async', true);
    $recommendedInterval = calculate_recommended_interval_user_activity($load_average, $batch_size = 75);
    wp_localize_script('heartbeat', 'dfehcData', array(
        'recommendedInterval' => $recommendedInterval,
    ));
}
add_action('wp_enqueue_scripts', 'dfehc_enqueue_scripts');

function calculate_recommended_interval($time_elapsed, $load_average, $server_response_time) {
    $min_interval = get_option('DFEHC_MIN_INTERVAL', DFEHC_MIN_INTERVAL);
    $max_interval = get_option('DFEHC_MAX_INTERVAL', DFEHC_MAX_INTERVAL);

    $user_activity_weight = 0.4;
    $server_load_weight = 0.3;
    $response_time_weight = 0.3;
    $ema_smoothing_factor = 0.4; 

    $user_activity_factor = $time_elapsed / $max_interval;
    $server_load_factor = 1 - ($load_average / DFEHC_MAX_SERVER_LOAD);
    $response_time_factor = $server_response_time / DFEHC_MAX_RESPONSE_TIME;

    $interval = $min_interval + (($user_activity_factor * $user_activity_weight +
        $server_load_factor * $server_load_weight +
        $response_time_factor * $response_time_weight) *
        ($max_interval - $min_interval));
    $previous_intervals = get_transient('dfehc_previous_intervals') ?: array();
    array_unshift($previous_intervals, $interval); 

    $ema = array_reduce($previous_intervals, function($carry, $item) use ($ema_smoothing_factor) {
        return $ema_smoothing_factor * $item + (1 - $ema_smoothing_factor) * $carry;
    }, $interval);  

    set_transient('dfehc_previous_intervals', $previous_intervals, 20 * MINUTE_IN_SECONDS);
    return $ema;
}

function calculate_recommended_interval_user_activity($load_average, $batch_size = 75) {
    $min_interval = get_option('DFEHC_MIN_INTERVAL', DFEHC_MIN_INTERVAL);
    $max_interval = get_option('DFEHC_MAX_INTERVAL', DFEHC_MAX_INTERVAL);

    $total_weighted_duration = 0;
    $total_weight = 0;

    $process_user_batch = function($users) use (&$total_weighted_duration, &$total_weight) {
        foreach ($users as $user) {
            $activity_data = get_user_meta($user->ID, 'dfehc_user_activity', true);
            if (empty($activity_data['durations'])) {
                continue;
            }
            $user_weight = count($activity_data['durations']);
            $user_average_duration = array_sum($activity_data['durations']) / count($activity_data['durations']);
            $total_weighted_duration += $user_weight * $user_average_duration;
            $total_weight += $user_weight;
        }
    };

    get_users_in_batches($batch_size, $process_user_batch);

    if ($total_weight === 0) {
       
        return $recommendedInterval;  
    }

    $average_duration = $total_weighted_duration / $total_weight;

    if ($average_duration <= $min_interval) {
        $interval = $min_interval;
    } else if ($average_duration >= $max_interval) {
        $interval = $max_interval;
    } else {
        $interval = calculate_recommended_interval($average_duration, $load_average, 0);
    }

    return $interval;
}

function dfehc_register_settings() {
    add_settings_section('dfehc_redis_settings_section', 'Redis Settings', 'dfehc_redis_settings_section_callback', 'dfehc_plugin');
    add_settings_field('dfehc_redis_server', 'Redis Server', 'dfehc_redis_server_callback', 'dfehc_plugin', 'dfehc_redis_settings_section');
    add_settings_field('dfehc_redis_port', 'Redis Port', 'dfehc_redis_port_callback', 'dfehc_plugin', 'dfehc_redis_settings_section');
    add_settings_field('dfehc_redis_socket', 'Redis Unix Socket', 'dfehc_redis_socket_callback', 'dfehc_plugin', 'dfehc_redis_settings_section');

    add_settings_section('dfehc_memcached_settings_section', 'Memcached Settings', 'dfehc_memcached_settings_section_callback', 'dfehc_plugin');
    add_settings_field('dfehc_memcached_server', 'Memcached Server', 'dfehc_memcached_server_callback', 'dfehc_plugin', 'dfehc_memcached_settings_section');
    add_settings_field('dfehc_memcached_port', 'Memcached Port', 'dfehc_memcached_port_callback', 'dfehc_plugin', 'dfehc_memcached_settings_section');

    register_setting('dfehc_options_group', 'dfehc_redis_server', 'dfehc_validate_server');
    register_setting('dfehc_options_group', 'dfehc_redis_port', 'dfehc_validate_port');
    register_setting('dfehc_options_group', 'dfehc_memcached_server', 'dfehc_validate_server');
    register_setting('dfehc_options_group', 'dfehc_memcached_port', 'dfehc_validate_port');
	register_setting('dfehc_options_group', 'dfehc_redis_socket');
}

function dfehc_get_redis_server() {
    return get_option('dfehc_redis_server', '127.0.0.1');
}

function dfehc_get_redis_port() {
    return get_option('dfehc_redis_port', 6379);
}

function dfehc_get_memcached_server() {
    return get_option('dfehc_memcached_server', '127.0.0.1');
}

function dfehc_get_memcached_port() {
    return get_option('dfehc_memcached_port', 11211);
}
function dfehc_redis_server_callback() {
    $server = get_option('dfehc_redis_server', '127.0.0.1');
    echo '<input type="text" name="dfehc_redis_server" value="' . esc_attr($server) . '" />';
}

function dfehc_redis_port_callback() {
    $port = get_option('dfehc_redis_port', 6379);
    echo '<input type="number" name="dfehc_redis_port" value="' . esc_attr($port) . '" />';
}

function dfehc_memcached_server_callback() {
    $server = get_option('dfehc_memcached_server', '127.0.0.1');
    echo '<input type="text" name="dfehc_memcached_server" value="' . esc_attr($server) . '" />';
}

function dfehc_memcached_port_callback() {
    $port = get_option('dfehc_memcached_port', 11211);
    echo '<input type="number" name="dfehc_memcached_port" value="' . esc_attr($port) . '" />';
}

function dfehc_redis_settings_section_callback() {
    echo '<p>Configure Redis settings for the plugin.</p>';
}

function dfehc_memcached_settings_section_callback() {
    echo '<p>Configure Memcached settings for the plugin.</p>';
}
function dfehc_redis_socket_callback() {
    $socket = get_option('dfehc_redis_socket', '/path/to/redis.sock');
    echo '<input type="text" name="dfehc_redis_socket" value="' . esc_attr($socket) . '" />';
}
function dfehc_add_settings_page() {
    add_options_page('DFEHC Settings', 'DFEHC', 'manage_options', 'dfehc_plugin', 'dfehc_settings_page_callback');
}

function dfehc_settings_page_callback() {
    ?>
    <div class="wrap">
        <h1>DFEHC Settings</h1>
		<p>If the default values don't work, only update the values of the main caching method you want to use. The plugin will automatically choose your enabled method. If you have multiple methods enabled, it's best you stick to just 1 main caching method. Redis is highly recommended in terms of speed.</p>
        <form action="options.php" method="post">
            <?php
            settings_fields('dfehc_options_group');
            do_settings_sections('dfehc_plugin');
            submit_button();
            ?>
        </form>
    </div>
    <?php
}

add_action('admin_menu', 'dfehc_add_settings_page');
add_action('admin_init', 'dfehc_register_settings');

global $dfehc_redis_available;
global $dfehc_memcached_available;

$dfehc_redis_available = class_exists('Redis');
$dfehc_memcached_available = class_exists('Memcached');

function dfehc_get_server_load_persistent() {
    global $dfehc_redis_available, $dfehc_memcached_available;

    $server_load = null;

    try {
        if ($dfehc_redis_available) {
            $redis = new Redis();
            $redis_connected = $redis->connect(dfehc_get_redis_server(), dfehc_get_redis_port());
            if ($redis_connected) {
                $server_load = $redis->get('server_load');
            }
            $redis->close();
        } elseif ($dfehc_memcached_available) {
            $memcached = new Memcached();
            $memcached_connected = $memcached->addServer(dfehc_get_memcached_server(), dfehc_get_memcached_port());
            if ($memcached_connected) {
                $server_load = $memcached->get('server_load');
            }
            $memcached->quit();
        }
    } catch (Exception $e) {
        error_log('DFEHC Error: ' . $e->getMessage());
    }

    return $server_load;
}

function dfehc_get_server_load_ajax() {
    $nonce = isset($_POST['nonce']) ? $_POST['nonce'] : '';
    $server_load = isset($_POST['serverLoad']) ? floatval($_POST['serverLoad']) : 0;
    $server_response_time = isset($_POST['serverResponseTime']) ? intval($_POST['serverResponseTime']) : 0;

    if (!wp_verify_nonce($nonce, 'dfehc-ajax-nonce')) {
        wp_send_json_error('Invalid nonce.');
    }
    if (function_exists('sys_getloadavg')) {
        $load = sys_getloadavg();
        $serverLoad = $load[0];
    } elseif (function_exists('exec') && !empty($_SERVER['SERVER_ADDR'])) {
        $serverIP = $_SERVER['SERVER_ADDR'];
        $command = "uptime";
        $output = [];
        exec($command, $output);
        preg_match('/load average: ([0-9.]+)/', implode(" ", $output), $matches);
        if (!empty($matches[1])) {
            $serverLoad = (float) $matches[1];
        }
    } elseif (is_readable('/proc/loadavg')) {
        $loadAvgFile = file_get_contents('/proc/loadavg');
        $loadAvgData = explode(' ', $loadAvgFile);
        $serverLoad = (float) $loadAvgData[0];
    } elseif (function_exists('shell_exec')) {
        $command = "uptime";
        $output = shell_exec($command);
        preg_match('/load average: ([0-9.]+)/', $output, $matches);
        if (!empty($matches[1])) {
            $serverLoad = (float) $matches[1];
        }
    } elseif (is_readable('/proc/stat')) {
        $statData = file_get_contents('/proc/stat');
        preg_match('/^cpu\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)/', $statData, $matches);

        if (isset($matches[1], $matches[2], $matches[3], $matches[4])) {
            $idle = $matches[4];
            $total = $matches[1] + $matches[2] + $matches[3] + $matches[4];
            $serverLoad = (1 - ($idle / $total)) * 100;
        }
    } elseif (function_exists('microtime')) {
        $startTime = microtime(true);
        $endTime = $startTime + 0.1;
        $busyTime = 0;
        while (microtime(true) < $endTime) {
            $busyTime++;
        }
        $idleTime = ($endTime - $startTime) - $busyTime;
        $serverLoad = ($busyTime / ($busyTime + $idleTime)) * 100;
    }
    $interval = calculate_recommended_interval_user_activity($server_load);

    if ($interval > 0) {
        wp_send_json_success($interval);
    } else {
        wp_send_json_error('Failed to calculate interval.');
    }
}
add_action('wp_ajax_get_server_load', 'dfehc_get_server_load_ajax');
add_action('wp_ajax_nopriv_get_server_load', 'dfehc_get_server_load_ajax');

function dfehc_get_recommended_heartbeat_interval_async() {
    if (!class_exists('Heartbeat_Async')) {
        include_once('heartbeat-async.php');
    }

    class Dfehc_Get_Recommended_Heartbeat_Interval_Async extends Heartbeat_Async {
        protected $action = 'dfehc_get_recommended_interval_async';

        protected function run_action() {
            $last_activity = (int) get_transient('dfehc_last_user_activity');
            $time_elapsed = time() - $last_activity;
            $load_average = sys_getloadavg()[0];

            $interval = calculate_recommended_interval($time_elapsed, $load_average, 0);

            set_transient('dfehc_recommended_interval', $interval, 5 * MINUTE_IN_SECONDS);
        }
    }

    $async_task = new Dfehc_Get_Recommended_Heartbeat_Interval_Async();
    $async_task->dispatch();

    return get_transient('dfehc_recommended_interval');
}

function dfehc_get_recommended_intervals() {
    check_ajax_referer('dfehc_get_recommended_intervals', 'nonce');

    $interval = dfehc_get_recommended_heartbeat_interval_async();

    $settings = array('interval' => $interval);

    wp_send_json_success($settings);
    wp_die();
}
add_action('wp_ajax_dfehc_update_heartbeat_interval', 'dfehc_get_recommended_intervals');
add_action('wp_ajax_nopriv_dfehc_update_heartbeat_interval', 'dfehc_get_recommended_intervals');

function dfehc_override_heartbeat_interval($settings) {
    $interval = isset($settings['interval']) ? $settings['interval'] : DFEHC_MIN_INTERVAL;

    if ($interval < DFEHC_MIN_INTERVAL) {
        $interval = DFEHC_MIN_INTERVAL;
    } else if ($interval > DFEHC_MAX_INTERVAL) {
        $interval = DFEHC_MAX_INTERVAL;
    }

    $settings['interval'] = $interval;

    return $settings;
}
add_filter('heartbeat_settings', 'dfehc_override_heartbeat_interval');

function dfehc_set_user_cookie() {
	 if (headers_sent()) {
        return;
    }
    $visitor_id = isset($_COOKIE['dfehc_user']) ? $_COOKIE['dfehc_user'] : uniqid('visitor_', true);
    setcookie('dfehc_user', $visitor_id, time() + 500, '/');

    $redis_available = extension_loaded('redis');
    $memcached_available = extension_loaded('memcached');
    $fallback_used = true;

    try {
        if ($redis_available) {
            $redis = new Redis();
            $redis_socket = get_option('dfehc_redis_socket', '/path/to/redis.sock');
            $connected = false;
            if (!empty($redis_socket)) {
                $connected = @$redis->connect($redis_socket);
            } else {
                $redis_server = dfehc_get_redis_server();
                $redis_port = dfehc_get_redis_port();
                $connected = @$redis->connect($redis_server, $redis_port);
            }
            
            if ($connected) {
                $redis->incr('dfehc_total_visitors');
                $redis->close();
                $fallback_used = false;
            }
        }

        if ($memcached_available && $fallback_used) {
            $memcached = new Memcached();
            $connected = @$memcached->addServer(dfehc_get_memcached_server(), dfehc_get_memcached_port());
            if ($connected) {
                $count = $memcached->get('dfehc_total_visitors');
                $memcached->set('dfehc_total_visitors', ($count ? $count : 0) + 1);
                $memcached->quit();
                $fallback_used = false;
            }
        }

        if ($fallback_used) {
            $count = get_transient('dfehc_total_visitors');
            set_transient('dfehc_total_visitors', ($count ? $count : 0) + 1);
        }
    } catch (Exception $e) {
        error_log('DFEHC Error in dfehc_set_user_cookie: ' . $e->getMessage());
        error_log('Redis server: ' . dfehc_get_redis_server() . ', Port: ' . dfehc_get_redis_port());
        error_log('Memcached server: ' . dfehc_get_memcached_server() . ', Port: ' . dfehc_get_memcached_port());
        error_log('Fallback used: ' . ($fallback_used ? 'Yes' : 'No'));
        error_log('Fallback count: ' . get_transient('dfehc_total_visitors'));
    }
}

add_action('init', 'dfehc_set_user_cookie');

function increment_total_visitors_fallback() {
    $count = get_transient('dfehc_total_visitors');
    set_transient('dfehc_total_visitors', ($count ? $count : 0) + 1);
}

function dfehc_get_website_visitors() {
    try {
        if (extension_loaded('redis')) {
            $redis = new Redis();
            $connected = $redis->connect(dfehc_get_redis_server(), dfehc_get_redis_port());
            if ($connected) {
                $result = $redis->get('dfehc_total_visitors');
            }
            $redis->close();
        } elseif (extension_loaded('memcached')) {
            $memcached = new Memcached();
            $connected = $memcached->addServer(dfehc_get_memcached_server(), dfehc_get_memcached_port());
            if ($connected) {
                $result = $memcached->get('dfehc_total_visitors');
            }
            $memcached->quit();
        } else {
            $result = get_transient('dfehc_total_visitors');
        }
    } catch (Exception $e) {
        error_log('DFEHC Error in dfehc_get_website_visitors: ' . $e->getMessage());
        $result = false;
    }

    return $result ?: 0;

}

function dfehc_invalidate_heartbeat_cache($user_id) {
    $high_visitor_count = get_transient('dfehc_high_visitor_count');
    if (!$high_visitor_count) {
        $visitor_count = dfehc_get_website_visitors();
        if ($visitor_count > 100) {
            set_transient('dfehc_high_visitor_count', true, 10 * MINUTE_IN_SECONDS);
            delete_transient('dfehc_recommended_interval');
        }
    } else {
        $last_recalculation = get_transient('dfehc_last_recalculation');

        if (!$last_recalculation || time() - $last_recalculation >= 10 * MINUTE_IN_SECONDS) {
            set_transient('dfehc_last_recalculation', time(), 10 * MINUTE_IN_SECONDS);
            delete_transient('dfehc_recommended_interval');
        }
    }
}
add_action('wp_logout', 'dfehc_invalidate_heartbeat_cache');
add_filter('heartbeat_received', 'dfehc_track_and_expire_user_activity', 10, 2);