(function($, wp) {
  const cacheTimeout = 5 * 60 * 1000;
  let cache = {};

  async function getRecommendedIntervals($nonce) {
    const cacheKey = 'dfehc_server_load';
    const cachedData = localStorage.getItem(cacheKey);
    const currentTime = new Date().getTime();

    if (cachedData) {
      const { timestamp, serverLoad } = JSON.parse(cachedData);
      if (currentTime - timestamp < cacheTimeout) {
        return serverLoad;
      }
    }

    try {
      const response = await $.ajax({
        url: ajaxurl,
        type: 'POST',
        data: {
          action: 'get_server_load',
          nonce: $nonce.val()
        }
      });

      if (response.success) {
        const serverLoad = response.data;
        localStorage.setItem(cacheKey, JSON.stringify({ timestamp: currentTime, serverLoad }));
        return serverLoad;
      } else {
        throw new Error(response.data.message);
      }

    } catch (error) {
      throw new Error(`Error retrieving server load: ${error}`);
    }
  }

  function smoothMoving(x) {
    let sum = 0;
    let y = [];
    for (let i = 0; i < x.length; i++) {
        if (y.length >= 5) {
            sum -= y.shift();
        }
        y.push(x[i]);
        sum += x[i];
    }
    return sum / y.length;
  }

  function getWeightedAverage(recentIntervals) {
    return recentIntervals.reduce((total, value, index) => {
        return total + value * (5 - index);
    }, 0) / 15;
  }

  function getTrafficLevel(serverLoad) {
    if (serverLoad <= 50) {
      return 'low';
    } else if (serverLoad <= 75) {
      return 'medium';
    } else {
      return 'high';
    }
  }

  function calculateRecommendedInterval(serverLoad) {
    if (cache[serverLoad]) {
        return cache[serverLoad];
    }

    const intervals = {
        'low': [15, 30, 60, 120, 180, 240, 300],
        'medium': [30, 60, 120, 180, 240, 300],
        'high': [60, 120, 180, 240, 300]
    };
    let trafficLevel = getTrafficLevel(serverLoad);
    let recommendedInterval;

    if (serverLoad !== null) {
        const recentIntervals = intervals[trafficLevel].slice(-5);
        const weightedAverage = getWeightedAverage(recentIntervals);
        const smoothedInterval = smoothMoving(recentIntervals);
        const maxServerLoad = 85;
        const loadFactor = 1 - (serverLoad / maxServerLoad);
        recommendedInterval = Math.round(weightedAverage + smoothedInterval * loadFactor);
    } else {
        const recentIntervals = intervals[trafficLevel].slice(-10);
        const sum = recentIntervals.reduce((a, b) => a + b, 0);
        recommendedInterval = Math.round(sum / recentIntervals.length + 5);
    }
	
    cache[serverLoad] = recommendedInterval;

    return recommendedInterval;
  }

  const heartbeat = {
    updateHeartbeatInterval: (interval) => wp.heartbeat.interval(interval),

    updateUI: function($heartbeatAutoSave, recommendedInterval, intervals) {
      const $intervalSelect = $('<select>').attr('id', 'dfehc-heartbeat-interval');
      intervals.forEach(interval => {
        const $option = $('<option>').attr('value', interval).text(`${interval} seconds`);
        $intervalSelect.append($option);
      });

      $heartbeatAutoSave.append($intervalSelect);
      this.updateHeartbeatInterval(recommendedInterval);
      $intervalSelect.val(recommendedInterval);
    },

    init: async function($heartbeatAutoSave, $nonce) {
      try {
        const serverLoad = await getRecommendedIntervals($nonce);
        const trafficLevel = getTrafficLevel(serverLoad);
        let clientRecommendedInterval = calculateRecommendedInterval(serverLoad);
        const serverRecommendedInterval = dfehcData.recommendedInterval;
        const serverWeight = 0.7;
        const clientWeight = 0.3;
        let recommendedInterval = Math.round((clientRecommendedInterval * clientWeight + serverRecommendedInterval * serverWeight) / (clientWeight + serverWeight));
        this.updateUI($heartbeatAutoSave, recommendedInterval, intervals);
      } catch (error) {
        console.error(`Error: ${error.message}`);
      }
    },
  };

  $(document).ready(function() {
    const $heartbeatAutoSave = $('#heartbeat-auto-save');
    const $nonce = $('#dfehc_vars_nonce');
    heartbeat.init($heartbeatAutoSave, $nonce);

    $heartbeatAutoSave.on('change', '#dfehc-heartbeat-interval', function() {
      const newInterval = $(this).val();
      heartbeat.updateHeartbeatInterval(newInterval);
    });
  });

})(jQuery, wp);